import logging
from pathlib import Path
from typing import Any, List, Optional, cast

# noinspection PyPackageRequirements
import rdkit

# noinspection PyPackageRequirements
import rdkit.Chem.AllChem

# noinspection PyPackageRequirements
import rdkit.Chem.Draw

# noinspection PyPackageRequirements
import rdkit.Chem.rdmolfiles
from kallisto.atom import Atom
from kallisto.molecule import Molecule
from kallisto.units import Bohr

from icb import constant
from icb.helper.bash import change_directory
from icb.types import T_RDKitMolecule

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def convert_rdkit_molecule_to_plain_xyz_data(molecule: T_RDKitMolecule) -> List[str]:
    # get all xyz coordinates and split into list of lines
    xyz: List[str] = rdkit.Chem.rdmolfiles.MolToXYZBlock(molecule).split("\n")
    # remove number of atoms and eventual molecule name
    xyz = xyz[2:]
    # remove empty lines from list
    xyz = [line for line in xyz if len(line.strip()) > 0]
    return xyz


def convert_rdkit_molecule_to_smiles(molecule: T_RDKitMolecule) -> str:
    return cast(str, rdkit.Chem.MolToSmiles(molecule))


def copy_rdkit_molecule(molecule: T_RDKitMolecule) -> T_RDKitMolecule:
    return rdkit.Chem.Mol(molecule)


def draw_rdkit_molecule_as_svg(molecule: T_RDKitMolecule, width: int, height: int) -> Any:
    svg = rdkit.Chem.Draw.rdMolDraw2D.MolDraw2DSVG(width, height)
    svg.DrawMolecule(molecule)
    svg.FinishDrawing()

    return svg


def load_mdl_file_as_rdkit_molecule(file_path: Path) -> T_RDKitMolecule:
    if file_path.suffix in constant.SUFFIX_MOL + constant.SUFFIX_SDF:
        return rdkit.Chem.MolFromMolFile(f"{file_path}")
    elif file_path.suffix in constant.SUFFIX_SDF:
        return rdkit.Chem.SDMolSupplier(f"{file_path}")
    else:
        msg = f"unsupported file type {file_path.suffix}, supported {constant.SUFFIX_MOL + constant.SUFFIX_SDF}"
        if LOG.isEnabledFor(logging.ERROR):
            LOG.error(msg)
        raise ValueError(msg)


def load_smiles_as_rdkit_molecule(smiles: str, sanitize: bool = True) -> T_RDKitMolecule:
    return rdkit.Chem.MolFromSmiles(smiles, sanitize=sanitize)


def rdkit_molecule_to_xyz(
    xyz_name: str,
    molecule: T_RDKitMolecule,
    path: Optional[Path] = None,
) -> Path:
    """Create xmol file from SMILES. Return path to xyz file."""

    _path = path if path else Path.cwd()

    source = Path.cwd()
    rdkit_molecule = copy_rdkit_molecule(molecule=molecule)

    change_directory(_path)
    molecule_with_hydrogen = rdkit.Chem.AddHs(rdkit_molecule)
    rdkit.Chem.AllChem.EmbedMolecule(molecule_with_hydrogen)
    plain_xyz = convert_rdkit_molecule_to_plain_xyz_data(molecule=molecule_with_hydrogen)

    pt = rdkit.Chem.GetPeriodicTable()

    # create atom list
    atoms = []
    for coord in plain_xyz:
        elem, x, y, z = coord.split()[:4]
        position = [float(x) / Bohr, float(y) / Bohr, float(z) / Bohr]
        atom = Atom(symbol=pt.GetAtomicNumber(elem), position=position)
        atoms.append(atom)
    # construct kallisto molecule
    kallisto_molecule = Molecule(symbols=atoms)
    # write xmol file
    kallisto_molecule.writeMolecule(name=xyz_name, path=_path)

    change_directory(source)

    return _path / xyz_name


import logging
from pathlib import Path
from typing import List

import click
import kallisto.reader.strucreader as ksr
from kallisto.rmsd import exchangeSubstructure

from icb.helper.bash import change_directory, copy_file, create_directory
from icb.helper.ts.store import write_transition_state
from icb.helper.xtb_helper import write_xtb_constraints

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_transition_states(
    atoms: List[int], index: int, jobid: str, templates: List[int]
) -> None:
    """Create transition-states and exchange substrates with kallisto"""

    if LOG.isEnabledFor(logging.INFO):
        LOG.info(f"Value of ch_positions {atoms} and type {type(atoms)}")

    cwd = Path.cwd()

    # extract atom by index
    atom = atoms[index]

    # build path to file
    path = Path.cwd() / f"{jobid}" / f"{atom}"
    inp = path / f"{atom}_wo_sorted.xyz"

    # define name of newstructure
    name = "newstructure.xyz"
    for template in templates:
        # create directory for template
        destination = path / f"{template}"
        create_directory(path=destination)
        # directory name for rotated substrates
        rotated = f"{template}r"
        rotated_destination = path / rotated
        # create directory for rotated substrate
        create_directory(path=rotated_destination)
        ts_name = f"ts{template}.xyz"
        write_transition_state(n=template, name=ts_name, path=destination)
        write_transition_state(n=template, name=ts_name, path=rotated_destination)

        # copy sorted substrate to transition-state directory
        copy_file(source=inp, destination=destination)
        copy_file(source=inp, destination=rotated_destination)

        # change to transition-state directory
        change_directory(path=destination)
        # create substrate object
        substrate = ksr.constructMolecule(geometry=f"{inp}", out=click.File())
        # number of atoms in substrate
        substrate_nat = substrate.get_number_of_atoms()
        # covalent bonding partners in substrate
        substrate_bonds = substrate.get_bonds()
        if LOG.isEnabledFor(logging.INFO):
            LOG.info(f"substrate={substrate}, inp={inp}, substrate_nat={substrate_nat}")
        # create transition-state template object
        ts = ksr.constructMolecule(geometry=ts_name, out=click.File())
        # number of atoms in template
        ts_nat = ts.get_number_of_atoms()
        # covalent bonding partners in template
        ts_bonds = ts.get_bonds()
        if LOG.isEnabledFor(logging.INFO):
            LOG.info(f"ts={ts}, ts_name={ts_name}, ts_nat={ts_nat}")

        # exchange benzene with substrate
        # benzene has substructure number 2
        benzene = 2
        # Iridium has index 18 in xmol
        iridium = 18
        mol = exchangeSubstructure(
            n=ts_nat,
            center=iridium,
            subnr=benzene,
            bonds=ts_bonds,
            ref=ts,
            newsub=substrate,
            newSubBonds=substrate_bonds,
            name="newstructure",
            rotate=0,
            exclude=False,
        )
        mol.writeMolecule(name=name, path=destination)
        write_xtb_constraints(index=template)
        if LOG.isEnabledFor(logging.INFO):
            LOG.info(f"mol.get_number_of_atoms()={mol.get_number_of_atoms()}")

        # change to transition-state directory with rotation
        change_directory(path=rotated_destination)

        # exchange benzene with substrate
        # and rotate substrate by 180 degrees
        # benzene has substructure number 2
        benzene = 2
        # Iridium has index 18 in xmol
        iridium = 18
        mol = exchangeSubstructure(
            n=ts_nat,
            center=iridium,
            subnr=benzene,
            bonds=ts_bonds,
            ref=ts,
            newsub=substrate,
            newSubBonds=substrate_bonds,
            name="newstructure",
            rotate=180,
            exclude=False,
        )
        mol.writeMolecule(name=name, path=rotated_destination)
        write_xtb_constraints(index=template)

        if LOG.isEnabledFor(logging.INFO):
            LOG.info(f"mol.get_number_of_atoms()={mol.get_number_of_atoms()}")
        # get back to working directory
        change_directory(path=cwd)

        # remove kallisto generated structure in wrong directory
        name_p = Path(name)
        if name_p.exists():
            name_p.unlink()


import logging
from pathlib import Path
from typing import List

from icb.helper.chem_helper import sort_xyz
from icb.helper.kallisto_helper import construct_molecule

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def sort_ch_with_breadth_first_search(
    atoms: List[int], index: int, jobid: str
) -> List[List[int]]:
    """Sort xmol structures according to breadth first search algorithm.
    List 'atoms' contains all C-H positions for which structures should be
    sorted. The number tells us which atom should be on zeroth position when
    The structure is sorted.

    Example:
     atoms = [2, 4]

     There exist two C-H positions in the molecule (2 and 4).
     There should exist already a directory 'positions', which incorporates one
     directory for each C-H position.

     Example from above:
      ./positions/2/2.xyz and ./positions/4/4.xyz

     Those structures MUST exist since they come from the first task in the DAGs definition.
     For structure ./position/2/2.xyz this present function will take the 2nd atom of the xmol file and
     will sort all entries such that the order will match a breadth first search order with respect
     to the connectivity of all atoms in the molecule."""

    if LOG.isEnabledFor(logging.INFO):
        LOG.info(f"Value of ch_positions {atoms} and type {type(atoms)}")

    # extract atom by index
    atom = atoms[index]

    # build path to files
    path = Path.cwd() / f"{jobid}" / f"{atom}"
    inp = path / "wo.xyz"
    out = f"{atom}_wo_sorted.xyz"
    print(path, inp)
    sort_xyz(inp=inp, start_atom=atom, out=out, path=path)

    # construct kallisto Molecule from sorted structure
    fdir = path / out
    kallisto_molecule = construct_molecule(geometry=fdir)
    graph = kallisto_molecule.get_bonds()

    return [sorted(x) for x in graph]


import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import joblib
import numpy as np
import pandas as pd

# noinspection PyPackageRequirements
from openbabel import pybel

# noinspection PyPackageRequirements
from rdkit import Chem

# noinspection PyPackageRequirements
from rdkit.Chem import AllChem
from scipy.spatial.distance import rogerstanimoto

from icb import utility
from icb.__init__ import __package__ as top_package_name
from icb.helper import rdkit_helper
from icb.helper.bash import copy_file, create_directory
from icb.helper.chem_helper import remove_atom_by_index
from icb.helper.ml.data import NUCLEAR_CHARGE
from icb.helper.utility import exist_file

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_machine_learning_deltas(
    atoms: List[int], jobid: str, templates: List[int], penalty: Dict[int, float]
) -> Tuple[Dict[str, float], Dict[str, float]]:
    """Create deltas from machine-learning model (random forest) for all C-H positions."""

    # define Compound object
    class Compound(object):
        """The Compound object is used to store data."""

        def __init__(self, xyz_file_path: Optional[Path] = None) -> None:

            empty_array = np.asarray([], dtype=float)

            self.molid = float("nan")
            self.xyz_file_path = xyz_file_path
            self.subsmile: Optional[str] = None

            # information about the compound
            self.nat = float("nan")
            self.at: Dict[str, int] = {}
            self.atomtypes: List[str] = []
            self.atomtype_indices: Dict[str, List[int]] = defaultdict(list)
            self.nuclear_charges = empty_array
            self.coordinates = empty_array
            self.natoms = -1

            # representation
            self.representation = empty_array

            self.read_xyz()

        def read_xyz(self) -> None:
            """
            (Re-)initializes the Compound-object with data from a xyz-file.

            """

            if self.xyz_file_path is None:
                return

            with self.xyz_file_path.open(mode="r") as f:
                lines = f.readlines()

            self.natoms = int(lines[0])
            self.atomtypes = []
            self.nuclear_charges = np.empty(self.natoms, dtype=int)
            self.coordinates = np.empty((self.natoms, 3), dtype=float)

            for ii, line in enumerate(lines[2 : self.natoms + 2]):
                tokens = line.split()

                if len(tokens) < 4:
                    break

                self.atomtypes.append(tokens[0])
                self.atomtype_indices[tokens[0]].append(ii)
                self.nuclear_charges[ii] = NUCLEAR_CHARGE[tokens[0]]

                self.coordinates[ii] = np.asarray(tokens[1:4], dtype=float)

            self.at = dict(
                [(key, len(value)) for key, value in self.atomtype_indices.items()]
            )

        def xyz_to_smiles(self) -> str:
            """Convert xyz file to smiles."""
            mol = next(pybel.readfile("xyz", f"{self.xyz_file_path}"))
            smiles = mol.write(format="smi")
            smile: str = smiles.split()[0].strip()
            return smile

        def get_representation(self, dimension: int = 128, size: int = 4) -> None:
            """Convert xyz to smile and calculate Morgan representation for cutout."""
            # extract substructure in complex via RDKit (turn off sanitization)
            rdk_molecule = rdkit_helper.load_smiles_as_rdkit_molecule(
                self.xyz_to_smiles(), sanitize=False
            )

            # set Carbon and Iridium nuclear charge
            carbon, iridium = 6, 77

            carbon_index = -1
            for atom_n in rdk_molecule.GetAtoms():
                # get Iridium atom index
                if atom_n.GetAtomicNum() == iridium:
                    # get neighbors of Iridium
                    neighbors = [x.GetAtomicNum() for x in atom_n.GetNeighbors()]
                    # sanity check: Carbon needs to be bound to Iridium
                    if carbon not in neighbors:
                        return
                    # Carbon position in neighbor list
                    carbon_position = neighbors.index(carbon)
                    # get indices of neighbors in SMILES
                    indices = [x.GetIdx() for x in atom_n.GetNeighbors()]
                    # get on-point: Carbon atom attached to Iridium
                    carbon_index = indices[carbon_position]

            # create sub_molecule in rdkit
            env = Chem.FindAtomEnvironmentOfRadiusN(rdk_molecule, size, carbon_index)
            sub_molecule = Chem.PathToSubmol(rdk_molecule, env)
            sub_molecule.UpdatePropertyCache(strict=False)
            Chem.SanitizeMol(
                sub_molecule,
                Chem.SanitizeFlags.SANITIZE_FINDRADICALS
                | Chem.SanitizeFlags.SANITIZE_KEKULIZE
                | Chem.SanitizeFlags.SANITIZE_SETAROMATICITY
                | Chem.SanitizeFlags.SANITIZE_SETCONJUGATION
                | Chem.SanitizeFlags.SANITIZE_SETHYBRIDIZATION
                | Chem.SanitizeFlags.SANITIZE_SYMMRINGS,
                catchErrors=True,
            )
            Chem.GetSymmSSSR(sub_molecule)
            # set ECFP4 fingerprint as representation
            representation = AllChem.GetMorganFingerprintAsBitVect(
                sub_molecule, 2, nBits=dimension
            )
            self.representation = np.array(representation)
            # set subsmile
            self.subsmile = Chem.MolToSmiles(sub_molecule)
            print(self.subsmile)

    # path specific
    positions = f"{jobid}"
    ml_dir_path = Path(positions) / "predict"

    # initialize base directory for ML prediction
    create_directory(path=ml_dir_path)

    # load machine-learning model
    pls_mod = "pls.mod-scikit-learn-1.1.1-python-3.10.0.joblib"
    model_file_path = utility.get_resource_path(top_package_name, f"data/{pls_mod}")
    print(f"Model path: {model_file_path}")
    if not model_file_path:
        raise FileNotFoundError("can not locate machine-learning model 'pls.mod' file")
    model = joblib.load(str(model_file_path))

    # load CSV data with binary training fingerprints
    db_file_path = utility.get_resource_path(top_package_name, "data/db.csv")
    if not db_file_path:
        raise FileNotFoundError(
            "can not locate binary training fingerprints 'db.csv' file"
        )
    df = pd.read_csv(db_file_path)
    df_list = df.values.tolist()

    # dimension of vector
    dim = 256

    # copy all structures
    deltas: Dict[str, float] = {}
    similarity: Dict[str, float] = {}
    for i, i_atom in enumerate(atoms):
        # save available templates in array
        # some templates might not be converged properly
        available: List[str] = []
        s_atom = f"{i_atom}"
        # initialize similarity
        similarity[s_atom] = 0
        # initialize delta
        deltas[s_atom] = 0
        ml_atom_dir_path = ml_dir_path / s_atom
        create_directory(path=ml_atom_dir_path)
        for i_template in templates:
            s_template = f"{i_template}"
            # create path to xmol file
            source = Path(positions) / s_atom / s_template / "xtbopt.xyz"
            destination = ml_atom_dir_path / f"{i_template}.xyz"
            # skip if optimization was not successful
            if exist_file(source):
                copy_file(source, destination)
                available.append(s_template)
            else:
                continue

        # create prediction vector
        prediction_vector = []
        similarity_list = []
        for s_template in available:
            # create path to xmol file
            destination = ml_atom_dir_path / f"{s_template}.xyz"
            # remove Hydrogen atom bound to Iridium
            # necessary since Iridium should not be hypervalent for openbabel
            remove_atom_by_index(index=19, inp=destination, out=destination)
            # create compound and representation
            cpd = Compound(xyz_file_path=destination)
            if LOG.isEnabledFor(logging.INFO):
                LOG.info(f"cpd.at={cpd.at}")

            # sanity check: compound creation
            if len(cpd.at) == 0:
                continue
            cpd.get_representation(dimension=dim)
            rep = cpd.representation
            # sanity check: representation creation
            if len(rep) == 0:
                continue
            prediction_vector.append(rep)

            # Rogers-Tanimoto similarity between vectors rep and row
            for j, row in enumerate(df_list):
                similarity_term = 1 - rogerstanimoto(rep, row)
                similarity_list.append(similarity_term)

        # extract similarity
        if similarity_list:
            similarity[s_atom] = max(similarity_list)

        # predict only when we have data
        if prediction_vector:
            # create numpy array
            np_x = np.array(prediction_vector)
            # predict deltas using randon forest
            yp = model.predict(np_x)
            # determine mean ML delta
            yp_mean = float(sum(yp) / len(yp))
            # sanity check
            # when predicted negative use graph penalty instead
            if yp_mean < -0.5:
                deltas[s_atom] = penalty[i_atom]
            else:
                # save mean delta to atomic position
                deltas[s_atom] = yp_mean
        else:
            deltas[s_atom] = 14

    return deltas, similarity


import logging
from typing import Any, Dict, List, Optional

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


class Results(object):
    """Results objects saves all results from the graph."""

    def __init__(
        self,
        molecular_similarity: int = 0,
        sort_ch: Optional[Any] = None,
        substrate_energy: float = float("nan"),
        templates: Optional[List[int]] = None,
        xtb_minimum: float = float("nan"),
    ) -> None:
        self.atoms: List[int] = []
        self.molecular_similarity: int = molecular_similarity
        self.substrate_energy: float = substrate_energy
        self.substrate_graph: Dict[int, List[List[int]]] = {}
        self.sort_ch: Optional[Any] = sort_ch
        self.templates: List[int] = templates if templates is not None else [1]
        self.xtb_energy: Dict[str, float] = {}  # optimize substrate and fix catalyst
        self.xtb_minimum: float = xtb_minimum
        self.xtb_gaps: Dict[str, float] = {}  # extract HOMO-LUMO gaps
        self.xtb_deltas: Dict[str, float] = {}
        self.xtb_ignore: Dict[int, List[int]] = {}  # xtb structures to be ignored
        self.similarity: Dict[int, float] = {}
        self.ml_deltas: Dict[int, float] = {}
        self.graph_deltas: Dict[int, float] = {}
        self.barriers: Dict[int, float] = {}
        self.weights: Dict[int, float] = {}


import logging

from icb.__version__ import __version__

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_header(smiles: str, name: str, cpus: str) -> str:
    """Create header of program."""
    return f"""
        Regioselectivity determination for the Iridium catalyzed borylation.
        Version {__version__}.

        Compound            : {smiles}
        Name                : {name}
        Number of CPU cores : {cpus}
    
    """


import logging
import os
from collections import defaultdict
from itertools import zip_longest
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

import click
import kallisto.reader.strucreader as ksr
from kallisto.data import chemical_symbols
from kallisto.units import Bohr

from icb.helper.utility import exist_file

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


class MolecularGraph:
    """Define a molecular graph.

    Methods:

     add_edge - adds edge to graph
     breadth_first_search_sort - breadth first search sorting

    Variables:

     inp (str) - input xmol structure
     out (str) - output xmol structure (breadth_first_search_sort sorted)
    """

    def __init__(self, inp: Path, out: Path):
        self.inp = inp
        self.out = out
        self.graph: Dict[int, List[int]] = defaultdict(list)
        self.molecule = ksr.constructMolecule(geometry=f"{self.inp}", out=click.File())
        self.nat = self.molecule.get_number_of_atoms()
        self.at = self.molecule.get_atomic_numbers()
        self.coordinates = self.molecule.get_positions()

    # method to add an edge to graph
    def add_edge(self, u: int, v: int) -> None:
        self.graph[u].append(v)

    # method to perform breadth first search
    def breadth_first_search_sort(self, s: int) -> None:

        # mark all vertices as not visited
        visited = [False] * (len(self.graph))

        # create a queue
        q = [s]

        # mark source node as visited and enqueue it
        visited[s] = True

        seperator = os.linesep
        with self.out.open(mode="w") as outFile:
            outFile.write(f"{self.nat}" + seperator)
            outFile.write("Created by Airflow" + seperator)
            while q:

                # dequeue a vertex from queue
                s = q.pop(0)
                outFile.write(
                    f"{chemical_symbols[self.at[s]]:3}"
                    f" {self.coordinates[s][0] * Bohr:9.4f}"
                    f" {self.coordinates[s][1] * Bohr:9.4f}"
                    f" {self.coordinates[s][2] * Bohr:9.4f}"
                    f"{seperator}"
                )

                # get adjacent vertices of dequeued vertex s
                # If an adjacent has not been visited, then mark
                # it as visited and enqueue it
                for i in self.graph[s]:
                    if visited[i] is False:
                        q.append(i)
                        visited[i] = True


def sort_xyz(inp: Path, start_atom: int, out: str, path: Optional[Path] = None) -> None:
    """Sort xyz structure according to connectivity matrix from kallisto program.
    Atom declared by 'start' will be the zeroth atom in the sorted structure."""

    _path = path if path else Path.cwd()

    # we need to create a dummy click output file to serve
    # the implemented kallisto API
    output_file = click.File(mode="r", lazy=True)
    kmol = ksr.constructMolecule(geometry=f"{inp}", out=output_file)

    # get number of atoms
    nat = kmol.get_number_of_atoms()
    bonds = kmol.get_bonds()

    # initialize molecular graph
    output_path = _path / out
    g = MolecularGraph(inp, output_path)

    # travel through the graph
    for i in range(nat):
        partners = bonds[i]
        for j in partners:
            # add edge (j, i) to graph
            g.add_edge(j, i)

    # breath first search sorting:
    #  start_atom will be zeroth atom in new structure
    #  output name defined by self.out
    g.breadth_first_search_sort(start_atom)


def remove_atom_by_index(
    index: int, inp: Path, out: Optional[Path] = None, path: Optional[Path] = None
) -> None:
    """Remove atom with index from structure."""

    _out = out if out else Path("wo.xyz")
    _path = path if path else Path.cwd()

    # check if input is available
    if not exist_file(inp):
        return

    location = _path / inp
    f = location.open(mode="r")
    lines = f.readlines()
    f.close()

    # reduce atom count
    nat = int(lines[0]) - 1
    # drop number of atoms line
    lines.pop(0)
    # increase index by 1
    index += 1
    # drop atom
    lines.pop(index)

    location = _path / _out
    f = location.open(mode="w")
    s = os.linesep
    f.write(f"{nat:>5}{s}")
    for line in lines:
        f.write(line)
    f.close()


def grouper(
    n: int, iterable: Iterable[Any], fillvalue: Optional[Any] = None
) -> Iterable[Tuple[Any, ...]]:
    """Collect data into fixed-length chunks or blocks."""
    # grouper(3, "ABCDEFG", "x") --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def extract_substrate(inp: Path, start: int, path: Optional[Path] = None) -> None:
    """Cut away atoms and extract substrate with nat atoms from catalyst structure."""

    _path = path if path else Path.cwd()

    with inp.open(mode="r+") as f:
        lines = f.readlines()

    nat = int(lines[0]) - start
    xyz = lines[2 + start :]
    dirp = _path / "submolecule.xyz"
    s = os.linesep
    with dirp.open(mode="w") as f:
        f.write(str(nat) + s)
        f.write(s)
        for coord in xyz:
            f.write(coord)


def get_atom_neighbors(
    atom_idx: int,
    molecule_covalent_nbrs: List[List[int]],
    alpha: Optional[List[int]] = None,
) -> Tuple[List[int], List[int], List[int]]:
    """Get all neighbors for atom_idx.

        Extract all covalent bonding partner (alpha), all nearest neighbours (beta),
    and all nearest-nearest neighbors (gamma) for atom with index 'atom_idx'.

    Args:
    atom_idx: index of atom to extract neighbors for
    molecule_covalent_list: list of covalent partner of each atom

    Returns:
    alpha: list of all covalent bonding atom indices of atom_idx
    beta: list of nearest neighbor atom indices of atom_idx
    gamma: list of nearest-nearest neighbor atom indices of atom_idx

    """
    # extract alpha neighbors
    effective_alpha: List[int] = alpha if alpha else molecule_covalent_nbrs[atom_idx]

    # extract beta neighbors
    beta: List[int] = []
    for _, a_alpha in enumerate(effective_alpha):
        b: List[int] = molecule_covalent_nbrs[a_alpha]
        diff = list({atom_idx} ^ set(b))
        if len(diff) > 0:
            beta.extend(diff)

    # extract gamma neighbors
    gamma: List[int] = []
    for _, a_beta in enumerate(beta):
        c: List[int] = molecule_covalent_nbrs[a_beta]
        inter = list(set(effective_alpha).intersection(set(c)))
        diff = list(set(inter) ^ set(c))
        gamma.extend(diff)
    gamma = list(dict.fromkeys(gamma))
    return effective_alpha, beta, gamma


import logging
from collections import defaultdict
from typing import Dict, List, cast

import numpy as np
from kallisto.atom import Atom
from kallisto.molecule import Molecule
from kallisto.rmsd import recursiveGetSubstructures
from kallisto.sterics import getClassicalSterimol
from kallisto.units import Bohr

# noinspection PyPackageRequirements
from rdkit import Chem

# noinspection PyPackageRequirements
from rdkit.Chem import AllChem, GetPeriodicTable

from icb.helper import rdkit_helper
from icb.helper.chem_helper import get_atom_neighbors
from icb.types import T_RDKitAtom, T_RDKitMolecule

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def atom_bound_to_excluded(atom: T_RDKitAtom) -> bool:
    """Returns if atoms is bounded to an excluded nuclear charge."""
    # exclude positions next to one of those nuclear charges
    # Nitrogen, Oxygen, and Sulfur
    excluded = [7, 8, 16]

    nbrs = atom.GetNeighbors()
    for nbr in nbrs:
        if nbr.GetAtomicNum() in excluded:
            return True
    return False


def create_graph_deltas(molecule: T_RDKitMolecule) -> Dict[int, float]:
    """Calculate the penalty depending on neighbors in molecular graph."""

    # copy RDKit molecule
    substrate = rdkit_helper.copy_rdkit_molecule(molecule)

    # identify aromatic Carbon atoms from pattern
    # and unpack tuple to get list of aromatic Carbon atoms
    pattern = Chem.MolFromSmarts("c")
    aromatic_carbon_positions = [
        atomPosition for (atomPosition,) in substrate.GetSubstructMatches(pattern)
    ]

    # get ring information and extract rings
    ri = substrate.GetRingInfo()
    rings = ri.AtomRings()

    # C-H count
    count = 0
    atoms = []

    # setup initial Hydrogen count and shift
    hcount, shift = -1, 0

    # Check for each atom
    for index, atom in enumerate(substrate.GetAtoms()):
        shift += 1
        hcount += atom.GetNumImplicitHs()
        hcount += atom.GetNumExplicitHs()
        for ring in rings:
            # extract Carbon atoms that
            # 1) belong to a ring system
            # 2) are aromatic
            # 3) occur only once (exclude doubles)
            # 4) have ONE C-H bond.
            if (
                (index in ring)
                and (index in aromatic_carbon_positions)
                and (index not in atoms)
                and (atom.GetNumImplicitHs() == 1)
            ):
                atoms.append(index)
                count += 1

    # add Hydrogen and embed
    substrate = Chem.AddHs(substrate)
    AllChem.EmbedMolecule(substrate)

    # build a kallisto molecule, get number of atoms
    # and molecular graph

    mk = kallisto_molecule_from_rdkit_molecule(substrate)
    # get coordinates of molecule
    molecule_coords = mk.get_positions()
    # get molecular graph
    molecular_graph = mk.get_bonds()
    # get atomic number of molecules
    atomic_numbers = mk.get_atomic_numbers()
    # get coordination numbers
    coordination_numbers = mk.get_cns(cntype="cov")
    # write xmol file (debug only)
    mk.writeMolecule("kallisto.xyz")
    # get number of atoms
    number_of_atoms = mk.get_number_of_atoms()

    # get atoms
    rdk_atoms = substrate.GetAtoms()

    # get nuclear charges
    atomic_numbers = [atom.GetAtomicNum() for atom in rdk_atoms]

    # define atoms that connect rings from dict
    atom_rings: Dict[int, int] = defaultdict(lambda: 0)
    for ring in rings:
        for atom in ring:
            atom_rings[atom] += 1
    # sort out atoms that are only in one ring and don't penalize position for excluded atom types
    connecting_ring_atoms: Dict[int, int] = dict(
        (k, v) for k, v in atom_rings.items() if v > 1
    )

    # get sorted rings with indices
    rings_sorted: Dict[int, List[int]] = defaultdict(lambda: [])
    for index, ring in enumerate(rings):
        rings_sorted[index] = sorted(ring)

    # iterate over all CHs
    sterical_penalties: Dict[int, Dict[int, float]] = {}

    for atom in atoms:
        # initialize
        sterical_penalties[atom] = {}

        # get RDKit atom object
        rdk_atom = rdk_atoms[atom]

        # extract neighbors of RDKit atom object
        neighbors = rdk_atom.GetNeighbors()

        # periodic table
        periodic_table = Chem.GetPeriodicTable()

        # iterate over neighbors
        for neighbor in neighbors:

            # sort out Hydrogen partners
            if neighbor.GetAtomicNum() != 1:

                # extract substructures using kallisto
                substructures = recursiveGetSubstructures(
                    number_of_atoms, molecular_graph, neighbor.GetIdx()
                )

                for substructure in substructures:
                    # sort out the ring substructures (see below)
                    # keep substituent only (has no 'atom' in substructure)
                    if atom not in substructure:
                        neighbor_covalent = [x.GetIdx() for x in neighbor.GetNeighbors()]
                        # tuple unpack single element set to int
                        (neighbor_neighbor,) = set(substructure).intersection(
                            set(neighbor_covalent)
                        )
                        # type cast to int is necessary
                        i_neighbor_neighbor = int(neighbor_neighbor)
                        # get nearest-nearest atom
                        neighbor_neighbor_atom = rdk_atoms[i_neighbor_neighbor]

                        # Only penalties for heavy atoms (not Hydrogen)
                        if rdk_atoms[i_neighbor_neighbor].GetAtomicNum() != 1:
                            # create molecule for modified Sterimol
                            subatoms = [
                                Atom(
                                    symbol=6,
                                    position=molecule_coords[neighbor.GetIdx()][:],
                                )
                            ]
                            for idx in substructure:
                                subatoms.append(
                                    Atom(
                                        symbol=atomic_numbers[idx],
                                        position=molecule_coords[idx][:],
                                    )
                                )
                            submolecule = Molecule(symbols=subatoms)
                            # write substituent (debug only)
                            # submolecule.writeMolecule("sub.xyz")
                            # Sterimol descriptors: L, Bmin, Bmax
                            l, bmin, bmax = getClassicalSterimol(submolecule, 0, 1)
                            if coordination_numbers[i_neighbor_neighbor] > 1.0:
                                # get Sterimol penalty from multivariant regression model
                                sterical_penalties[atom][
                                    neighbor.GetIdx()
                                ] = sterical_penalty(l, bmin, bmax)
                                # print(
                                #     'nbr, neighbor substituent, L, bmin, bmax, penalty',
                                #     int(nbr.GetIdx()), int(nn), l_value, bmin, bmax,
                                #     sterical_penalties[atom][nbr.GetIdx()], cns[nn], cns[nbr.GetIdx()]
                                # )
                            else:
                                sterical_penalties[atom][neighbor.GetIdx()] = 0

                            # Special cases
                            # increase penalty for atoms by sqrt(atom_mass)
                            if coordination_numbers[i_neighbor_neighbor] < 1.0:
                                print(
                                    sterical_penalties[atom][neighbor.GetIdx()],
                                    np.sqrt(neighbor_neighbor_atom.GetMass()),
                                )
                                sterical_penalties[atom][neighbor.GetIdx()] += np.sqrt(
                                    neighbor_neighbor_atom.GetMass()
                                )

    # create panalties
    graph_deltas: Dict[int, float] = {}
    for atom in atoms:
        graph_deltas[atom] = 0
        penalties = sterical_penalties[atom]
        for _, penalty in penalties.items():
            graph_deltas[atom] += penalty

    # now we construct penalties for atoms next to connected ring atoms
    atoms_connecting_rings = []
    for atom in atoms:
        sterical_penalties[atom] = {}
        # get RDKit atom object
        rdk_atom = rdk_atoms[atom]

        # extract neighbors of RDKit atom object
        neighbors = rdk_atom.GetNeighbors()

        # iterate over neighbors
        for neighbor in neighbors:

            # sort out Hydrogen partners
            if neighbor.GetAtomicNum() != 1:

                # neighbor to connecting ring atom
                __idx = [rdk_atom.GetIdx()]
                if neighbor.GetIdx() in connecting_ring_atoms:
                    atoms_connecting_rings.append(rdk_atom.GetIdx())
                    # print("in ring, connecting", nbr.GetIdx(), rdatom.GetIdx())
                    alpha, beta, gamma = get_atom_neighbors(
                        neighbor.GetIdx(), molecular_graph, [rdk_atom.GetIdx()]
                    )
                    subatoms = [
                        Atom(symbol=6, position=molecule_coords[neighbor.GetIdx()][:])
                    ]
                    for idx in alpha:
                        subatoms.extend(
                            [
                                Atom(
                                    symbol=atomic_numbers[idx],
                                    position=molecule_coords[idx][:],
                                )
                            ]
                        )
                    for idx in beta:
                        subatoms.extend(
                            [
                                Atom(
                                    symbol=atomic_numbers[idx],
                                    position=molecule_coords[idx][:],
                                )
                            ]
                        )
                    for idx in gamma:
                        subatoms.extend(
                            [
                                Atom(
                                    symbol=atomic_numbers[idx],
                                    position=molecule_coords[idx][:],
                                )
                            ]
                        )
                    submolecule = Molecule(symbols=subatoms)
                    # write submolecule (debug only)
                    # submolecule.writeMolecule("submolecule-2.xyz")
                    # Sterimol descriptors: L, Bmin, Bmax
                    l, bmin, bmax = getClassicalSterimol(submolecule, 0, 1)
                    # get a value penalty from multivariant regression model
                    sterical_penalties[atom][neighbor.GetIdx()] = sterical_penalty(
                        l, bmin, bmax
                    )

    # extend penalties with ortho to connecting ring atoms
    for atom in atoms_connecting_rings:
        penalties = sterical_penalties[atom]
        for _, penalty in penalties.items():
            graph_deltas[atom] += penalty
    return graph_deltas


def get_distance(origin: List[float], partner: List[float]) -> float:
    """Calculate the distance between origin and partner."""
    return cast(
        float,
        np.sqrt(
            np.power(origin[0] - partner[0], 2)
            + np.power(origin[1] - partner[1], 2)
            + np.power(origin[2] - partner[2], 2)
        ),
    )


def kallisto_molecule_from_rdkit_molecule(molecule: T_RDKitMolecule) -> Molecule:
    """Create a kallisto molecule from RDKit molecule.
    Args:
        molecule: RDKit molecule
    Returns:
        A kallisto molecule (kallisto.molecule.Molecule)
    """
    plain_xyz = rdkit_helper.convert_rdkit_molecule_to_plain_xyz_data(molecule=molecule)

    # setup periodic table
    pt = GetPeriodicTable()
    # create list of atoms
    atoms = []
    # create kallisto molecule
    for coord in plain_xyz:
        elem, x, y, z = coord.split()[:4]

        # convert atomic coordinates from Angstrom to Bohr
        position = [float(x) / Bohr, float(y) / Bohr, float(z) / Bohr]
        atom = Atom(symbol=pt.GetAtomicNumber(elem), position=position)
        atoms.append(atom)
    return Molecule(symbols=atoms)


def sterical_penalty(l_value: float, bmin: float, bmax: float) -> float:
    """Create penalties from Sterimol descriptors via a multivariant
    linear regression model fitted to B3-LYP/def2-TZVP penalties as obtained
    for SiMe3 probe groups in ortho position to the substituent."""

    intercept = 15.24292747
    l_s = -0.23004497
    bmin_s = -7.76699641
    bmax_s = 4.95959058
    penalty = intercept + l_s * l_value + bmin_s * bmin + bmax_s * bmax
    if penalty < 0.0:
        return float(0)
    return penalty


import logging
import os
from pathlib import Path
from typing import List, Optional

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())

# define global line seperator
LINE_SEP = os.linesep


def fix_atoms(atoms: List[int], file_path: Optional[Path] = None) -> None:
    """Write a xtb fixing file for passed in atoms."""

    _file_path = file_path if file_path else Path("fix.inp")

    # build string
    string = ", "
    string = string.join(map(str, atoms))

    with _file_path.open(mode="wt") as f:
        f.write(f"$fix{LINE_SEP} atoms: {string}{LINE_SEP}$end{LINE_SEP}")


def write_xtb_constraints(index: int) -> None:
    """Write xtb constraints depending on transition-state template index."""

    if index in [1, 2, 3, 4, 5, 6, 7, 8, 9]:
        # fix dihedral: B(63)-Ir(19)-H(20)-C(86)
        fix_atoms(atoms=[63, 19, 20, 86])
        return

    # fix dihedral: B(42)-Ir(19)-H(20)-C(86)
    fix_atoms(atoms=[42, 19, 20, 86])
    return


def calculate_barrier(ts: float, cat: float, sub: float, conv: float = 2625.5) -> float:
    """Calculate all barriers for C-H activation using energies for
    ts: transition-state energy in Hartree
    cat: catalyst energy in Hartree
    sub: substrate energy in Hartree.
    Returns by default a barrier in kJ/mol."""

    # default conversion factor for  Hartree -> kJ/mol
    return (ts - cat - sub) * conv


def extract_homo_lumo_gap(inp: Path) -> float:
    """Extract the homo-lumo gap from a xtb output."""
    try:
        with inp.open(mode="r+") as f:
            contents = f.readlines()
        target = "HOMO-LUMO GAP"
        targets = [ss for ss in contents if target in LINE_SEP]
        # return homo-lumo gap
        return float(targets[0].split()[3])
    except Exception:
        return float(0)


__all__ = [
    "DEFAULT_XTB_NUM_CPU_CORES_CONSTRAINT",
    "CSV_DECIMAL_CHAR",
    "CSV_SEP_CHAR",
    "NOK",
    "OK",
    "SUFFIX_MOL",
    "SUFFIX_SDF",
    "SUFFIX_SMILES",
    "XTB_EXEC_NAME",
    "XTB_EXEC_STDERR_FILE_NAME",
    "XTB_EXEC_STDOUT_FILE_NAME",
    "XTB_OPTIMISED_XYZ_FILE_NAME",
]

CSV_DECIMAL_CHAR = '.'
CSV_SEP_CHAR = ';'
DEFAULT_XTB_NUM_CPU_CORES_CONSTRAINT = 4

OK = 0
NOK = 1

SUFFIX_MOL = ['.mol']
SUFFIX_SDF = ['.sdf']
SUFFIX_SMILES = ['.smi', '.smiles']

XTB_EXEC_NAME = "xtb"
XTB_EXEC_STDERR_FILE_NAME = "xtb.stderr"
XTB_EXEC_STDOUT_FILE_NAME = "xtb.stdout"
XTB_OPTIMISED_XYZ_FILE_NAME = "xtbopt.xyz"


import importlib
import importlib.util
import logging
import os
import shutil
import sys
from pathlib import Path
from typing import List, Optional, Union

__all__ = [
    "change_working_directory",
    "copy_file",
    "create_non_existing_directory",
    "get_resource_path",
    "save_to_file",
]

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def change_working_directory(path: Path) -> None:
    """Change directory to path."""

    if LOG.isEnabledFor(logging.DEBUG):
        LOG.debug(f"path={path}")

    if path.exists() and path.is_dir():
        os.chdir(path)
    else:
        msg = f"the directory {path.absolute()} do not exist."
        if LOG.isEnabledFor(logging.ERROR):
            LOG.error(msg)
        raise FileNotFoundError(msg)


def copy_file(source: Path, destination: Path) -> None:
    """Copy file from source (path) to destination (path)."""

    if LOG.isEnabledFor(logging.DEBUG):
        LOG.debug(f"source={source}, destination={destination}")

    shutil.copy2(source, destination)


def create_non_existing_directory(name: str, path: Optional[Path] = None) -> Path:
    """Create directory for mentioned file_path."""

    if LOG.isEnabledFor(logging.DEBUG):
        LOG.debug(f"file_path={name}, path={path}")

    candidate_path = path / name if path else Path.cwd() / name

    if candidate_path.exists() and not candidate_path.is_dir():
        msg = f"the path {candidate_path.absolute()} exist and is not a directory."
        if LOG.isEnabledFor(logging.ERROR):
            LOG.error(msg)
        raise OSError(msg)

    if not candidate_path.exists():
        candidate_path.mkdir(parents=True)

    return candidate_path


def get_resource_path(package: str, resource: str) -> Optional[Path]:
    if LOG.isEnabledFor(logging.DEBUG):
        LOG.debug(f"package={package}, resource={resource}")

    spec = importlib.util.find_spec(package)
    if spec is None:
        return None
    mod = sys.modules.get(package)
    if mod is None or not hasattr(mod, "__file__") or not mod.__file__:
        return None

    parts: List[str] = resource.split(os.sep)
    parts.insert(0, os.path.dirname(mod.__file__))
    resource_name = os.path.join(*parts)

    return Path(resource_name)


def save_to_file(
    data: Optional[Union[bytes, str]] = None, file_path: Optional[Path] = None
) -> int:
    if LOG.isEnabledFor(logging.DEBUG):
        LOG.debug(f"data={data!r}, file_path={file_path!r}")

    num_written = 0
    if data is None:
        return num_written
    if file_path is None:
        return num_written

    m = 'wt' if isinstance(data, str) else 'wb'
    with file_path.open(mode=m) as f:
        num_written = f.write(data)

    return num_written


import logging
import shutil
import subprocess
from pathlib import Path
from typing import List, Optional, Text

from icb import constant, utility

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def extract_energy(input_file_path: Path) -> float:
    """Extract energy from optimized xtb structure."""
    if input_file_path.exists():
        with input_file_path.open(mode="rt") as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                if i == 1:
                    fields = line.split()
                    energy = fields[1]
                    break
        return float(energy)
    else:
        return float(14)


def perform_xtb_calculation(
    input_file_path: Path, args: List[str], dir_path: Path
) -> float:
    """Perform an xTB calculation.

    Inputs:
    input_file_path: Path to the input file (e.g., structures/input.xyz).
    args: Arguments defining the xTB calculation.
    dir_path: Name of temporary directory.

    """
    xtb_exec_path: Optional[str] = shutil.which(constant.XTB_EXEC_NAME)
    if not xtb_exec_path:
        msg = f"the {constant.XTB_EXEC_NAME} executable is not in PATH"
        if LOG.isEnabledFor(logging.CRITICAL):
            LOG.critical(msg=msg)
        raise OSError(msg)

    stderr_file_path = dir_path / constant.XTB_EXEC_STDERR_FILE_NAME
    stdout_file_path = dir_path / constant.XTB_EXEC_STDOUT_FILE_NAME

    args = [xtb_exec_path, input_file_path.name] + args
    exec_result: subprocess.CompletedProcess[Text] = subprocess.run(
        args,
        capture_output=True,
        cwd=f"{dir_path}",
        text=True,
    )

    if 0 != exec_result.returncode:
        msg = (
            f"failed during execution of {xtb_exec_path}"
            f", exit_code={exec_result.returncode}"
            f", stderr={exec_result.stderr}"
            f", stdout={exec_result.stdout}"
        )
        if LOG.isEnabledFor(logging.CRITICAL):
            LOG.critical(msg=msg)
        raise OSError(msg)

    utility.save_to_file(
        data=exec_result.stderr,
        file_path=stderr_file_path,
    )
    utility.save_to_file(
        data=exec_result.stdout,
        file_path=stdout_file_path,
    )

    return extract_energy(input_file_path=dir_path / constant.XTB_OPTIMISED_XYZ_FILE_NAME)


import logging
from pathlib import Path
from typing import IO, List, Text

import kallisto.atom
import kallisto.molecule
import kallisto.units

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def construct_molecule(geometry: Path) -> kallisto.molecule.Molecule:
    """Helper function to construct a Molecule."""
    try:
        with geometry.open(mode="r+") as fileObject:
            # read atoms from input file
            atoms = read_xyz(fileObject)
            # create molecule from atoms
            molecule = kallisto.molecule.Molecule(symbols=atoms)
    except FileNotFoundError:
        raise FileNotFoundError("Input file not found.")
    return molecule


def read_xyz(f: IO[Text]) -> List[kallisto.atom.Atom]:
    """Read xmol file and return list of atoms."""
    atoms = []
    lines = f.readlines()
    nat = int(lines[0])
    for line in lines[2 : nat + 2]:
        atom, x, y, z = line.split()[:4]
        symbol = atom.strip()[0].upper() + atom.strip()[1:].lower()
        position = [
            float(x) / kallisto.units.Bohr,
            float(y) / kallisto.units.Bohr,
            float(z) / kallisto.units.Bohr,
        ]
        atoms.append(kallisto.atom.Atom(symbol=symbol, position=position))
    return atoms


import logging
from pathlib import Path
from typing import Any, List

# noinspection PyPackageRequirements
from rdkit import Chem

from icb.helper import rdkit_helper
from icb.helper.bash import copy_file, create_directory
from icb.helper.chem_helper import remove_atom_by_index
from icb.types import T_RDKitAtom, T_RDKitMolecule

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_hydrogen_reduced_structure(jobid: str, molecule: T_RDKitMolecule) -> List[int]:
    """Identify all aromatic C-H positions from SMILES and
    cut away the Hydrogen atom attached to aromatic Carbon.
    Write substructures to directories.
    Return list of aromatic C-H atoms."""

    # path specific
    positions = f"{jobid}"

    # initialize base directory for method
    create_directory(path=Path(positions))

    # create substrate from input
    substrate = rdkit_helper.copy_rdkit_molecule(molecule=molecule)

    rdatoms = substrate.GetAtoms()

    # identify aromatic Carbon atoms from pattern
    # and unpack tuple to get list of aromatic Carbon atoms
    pattern: T_RDKitMolecule = Chem.MolFromSmarts("c")
    aromatic_carbon_positions = [
        atomPosition for (atomPosition,) in substrate.GetSubstructMatches(pattern)
    ]

    # get ring information and extract rings
    ri = substrate.GetRingInfo()
    rings = ri.AtomRings()

    # C-H candidates
    atoms: List[int] = []
    hydrogen_neighbors: List[int] = []

    # setup initial Hydrogen count and shift
    hcount = -1
    shift = 0

    nitrogen_nuclear_charge = 7
    # Check for each atom
    for index, atom in enumerate(substrate.GetAtoms()):
        shift += 1
        hcount += atom.GetNumImplicitHs()
        hcount += atom.GetNumExplicitHs()
        for ring in rings:
            #################################################
            # SPECIAL CASE for Nitrogen
            #################################################
            # sort out ortho to Nitrogen in six-membered ring
            if len(ring) == 6:
                if check_for_neighbor_atom(
                    atom=atom, nuclear_charge=nitrogen_nuclear_charge
                ):
                    continue
            # sort out ortho to Nitrogen in five-membered ring
            # ONLY when more than one Nitrogen is in ring
            elif len(ring) == 5:
                at = [rdatoms[atom].GetAtomicNum() for atom in ring]
                if check_for_atom_duplicates(at, nitrogen_nuclear_charge):
                    if check_for_neighbor_atom(
                        atom=atom, nuclear_charge=nitrogen_nuclear_charge
                    ):
                        continue
            #################################################
            # extract Carbon atoms that
            # 1) belong to a ring system
            # 2) are aromatic
            # 3) occur only once (exclude doubles)
            # 4) have ONE C-H bond.
            if (
                (index in ring)
                and (index in aromatic_carbon_positions)
                and (index not in atoms)
                and (atom.GetNumImplicitHs() == 1)
            ):
                atoms.append(index)
                hydrogen_neighbors.append(hcount)

    # Hydrogen atoms are added to the end of the xmol file
    # therefore shift all indices with respect to the number
    # of atoms that are not Hydrogen
    hydrogen_neighbors = [x + shift for x in hydrogen_neighbors]

    # prepare xmol structures
    xyz_name = "smiles.xyz"
    dirp = Path.cwd() / positions
    rdkit_helper.rdkit_molecule_to_xyz(xyz_name=xyz_name, molecule=molecule, path=dirp)
    # combine path and file name
    source = dirp / xyz_name
    for i, k in enumerate(atoms):
        path = dirp / f"{k}"
        create_directory(path=path)
        destination = path / f"{k}.xyz"
        copy_file(source, destination)
        # remove Hydrogen atom connected to aromatic Carbon
        remove_atom_by_index(index=hydrogen_neighbors[i], inp=destination, path=path)

    return atoms


def check_for_atom_duplicates(atoms: List[int], nuclear_charge: Any) -> bool:
    """Check for duplicates of nuclear_charges in atoms list."""
    # check if nuclear charge exists at all
    if nuclear_charge not in atoms:
        return False
    # set up a hash table to count occurrence
    hash_dict = {}
    for elem in atoms:
        if elem not in hash_dict:
            hash_dict[elem] = 1
        else:
            hash_dict[elem] += 1
    # nuclear charge is more than once in atom list
    if hash_dict[nuclear_charge] > 1:
        return True

    return False


def check_for_neighbor_atom(atom: T_RDKitAtom, nuclear_charge: float) -> bool:
    """Check if atom is next to a specific Atom (True) or not (False)."""

    # get all neighbors
    nbrs = atom.GetNeighbors()
    # iterate over neighbors

    for nbr in nbrs:
        if nbr.GetAtomicNum() == nuclear_charge:
            return True

    return False


#!/usr/bin/env python3
import logging
import os
import re

# arguments via ArgumentParser
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple

import pandas
import rich.table
import rich.text

# rich output formatter
from rich.console import Console
from rich.markdown import Markdown

from icb import constant

# helper function for structure validation, kallisto molecule and xtb gaps
from icb.helper import (
    chem_helper,
    kallisto_helper,
    rdkit_helper,
    utility,
    xtb_helper,
)

# helper functions
from icb.main.header import create_header

# results object
from icb.main.results import Results

# workflow specific tasks
from icb.task.create_barriers import create_barriers
from icb.task.create_boltzmann_weight_image import create_boltzmann_weight_image
from icb.task.create_ch import create_hydrogen_reduced_structure
from icb.task.create_graph_deltas import create_graph_deltas
from icb.task.create_ml_deltas import create_machine_learning_deltas
from icb.task.create_ts import create_transition_states
from icb.task.create_xtb_deltas import create_xtb_deltas
from icb.task.optimize_structure import perform_xtb_calculation
from icb.task.sort_ch import sort_ch_with_breadth_first_search
from icb.types import T_RDKitMolecule

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


# define the workflow
def main() -> int:
    #######################################################################
    # general setup
    #######################################################################

    # xtb optimization level
    xtb_opt_level = "lax"

    # input arguments: number of CPU's, name and SMILES
    parser = ArgumentParser()
    parser.add_argument(
        "-c",
        "--cpus",
        dest="num_cpu_cores",
        type=int,
        default=constant.DEFAULT_XTB_NUM_CPU_CORES_CONSTRAINT,
        help=(
            f"Number of CPU cores used during XTB calculations"
            f", default is {constant.DEFAULT_XTB_NUM_CPU_CORES_CONSTRAINT}."
        ),
    )
    parser.add_argument(
        "-i",
        "--input_file",
        dest="input_file",
        type=str,
        help=(
            f"Path to a input file."
            f" Supported formats are Daylight SMILES, and MDL MOL and SDFile."
            f" Override the --smiles option."
        ),
    )
    parser.add_argument(
        "-n",
        "--name",
        dest="name",
        type=str,
        help=(
            f"Used as working directory and report file stem."
            f" White spaces and '{os.sep}' characters are replaced by '_' characters."
        ),
    )
    parser.add_argument("-s", "--smiles", dest="smiles", type=str)
    args = parser.parse_args()

    # sanity checks for INPUT_FILE, NAME and SMILES values
    rdkit_molecule, smiles = _validate_args(parser=parser, args=args)

    # number of CPUs for xtb optimizations
    num_cpu_cores = args.num_cpu_cores

    # set start time
    start_time = datetime.now()

    #######################################################################
    # construct a rich console and create header
    #######################################################################
    console = Console()

    name = re.sub(os.sep, "_", re.sub(r"\s", "_", args.name))

    header = create_header(smiles, name, args.num_cpu_cores)
    md = Markdown(header)
    console.print(md)

    #######################################################################
    # workflow begin
    #######################################################################

    # initialize a results object
    res = Results()

    _create_all_ch_positions(
        console=console, results=res, name=name, molecule=rdkit_molecule
    )
    _optimize_substrate_with_gfn2_xtb_alpb_thf(
        console=console,
        results=res,
        name=name,
        num_cpu_cores=num_cpu_cores,
        xtb_opt_level=xtb_opt_level,
    )
    _sort_ch_positions(console=console, results=res, name=name)
    _create_all_transition_states(console=console, results=res, name=name)
    _optimize_transition_states_with_gfn2_xtb_alpb_thf(
        console=console,
        results=res,
        name=name,
        num_cpu_cores=num_cpu_cores,
        xtb_opt_level=xtb_opt_level,
    )
    _calculate_xtb_deltas_from_energies(console=console, results=res)
    _calculate_graph_deltas_from_molecular_graph(
        console=console,
        results=res,
        molecule=rdkit_molecule,
    )

    _calculate_machine_learning_deltas_using_ts_structures(
        console=console, results=res, name=name
    )

    _calculate_barriers_in_kilo_joule_per_mol_and_weights_in_percent(
        console=console, results=res
    )

    #######################################################################
    # workflow end
    #######################################################################

    console.print()
    end_time = datetime.now()

    rich_table = _print_properties_to_console(console, res)
    console.print()
    data = _rich_table_to_pandas_dataframe(
        table=rich_table,
    )
    _create_csv_report(
        console=console,
        data=data,
        name=name,
    )
    console.print()
    _create_molecular_image_report(
        console=console,
        results=res,
        name=name,
        molecule=rdkit_molecule,
    )

    console.print()
    console.print("Overall calculation time: ", end_time - start_time)

    return 0


def _calculate_barriers_in_kilo_joule_per_mol_and_weights_in_percent(
    console: Console, results: Results
) -> None:
    #######################################################################
    # calculate barriers in kJ/mol and weights in percent
    #######################################################################
    md = Markdown(" Final barriers and Boltzmann weights")
    console.print(md)
    step_begin = datetime.now()
    res_barriers, res_weights = create_barriers(
        atoms=results.atoms,
        minimum=results.xtb_minimum,
        xtb_deltas=results.xtb_deltas,
        ml_deltas=results.ml_deltas,
        graph_deltas=results.graph_deltas,
        similarity=results.similarity,
        molecular_similarity=results.molecular_similarity,
    )
    if res_barriers is not None:
        results.barriers = res_barriers
    if res_weights is not None:
        results.weights = res_weights
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _calculate_graph_deltas_from_molecular_graph(
    console: Console, results: Results, molecule: T_RDKitMolecule
) -> None:
    #######################################################################
    # calculate graph deltas from molecular graph
    #######################################################################
    md = Markdown(" Molecular graph penalties")
    console.print(md)
    step_begin = datetime.now()
    res_graph_deltas = create_graph_deltas(molecule=molecule)
    if res_graph_deltas is not None:
        results.graph_deltas = res_graph_deltas
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _calculate_machine_learning_deltas_using_ts_structures(
    console: Console, results: Results, name: str
) -> None:
    #######################################################################
    # calculate machine-learning deltas using TS structures
    #######################################################################
    md = Markdown(" PLS-P2 machine-learning penalties")
    console.print(md)
    step_begin = datetime.now()
    res_ml_deltas, res_similarity = create_machine_learning_deltas(
        atoms=results.atoms,
        jobid=name,
        templates=results.templates,
        penalty=results.graph_deltas,
    )
    if res_ml_deltas is not None:
        results.ml_deltas = {int(k): v for k, v in res_ml_deltas.items()}
    if res_similarity is not None:
        results.similarity = {int(k): v for k, v in res_similarity.items()}
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _calculate_xtb_deltas_from_energies(console: Console, results: Results) -> None:
    #######################################################################
    # calculate xtb deltas from energies
    #######################################################################
    md = Markdown(" GFN2-xtb/ALPB(THF) minimum and penalties")
    console.print(md)
    step_begin = datetime.now()
    res_xtb_minimum, res_xtb_deltas = create_xtb_deltas(
        atoms=results.atoms,
        templates=results.templates,
        substrate=results.substrate_energy,
        xtb_energy=results.xtb_energy,
        xtb_gaps=results.xtb_gaps,
        xtb_ignore=results.xtb_ignore,
    )
    if res_xtb_minimum is not None:
        results.xtb_minimum = res_xtb_minimum
    if res_xtb_deltas is not None:
        results.xtb_deltas = res_xtb_deltas
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _create_all_ch_positions(
    console: Console,
    results: Results,
    name: str,
    molecule: T_RDKitMolecule,
) -> None:
    #######################################################################
    # create all CH positions
    #######################################################################
    md = Markdown(" Generate all C-H positions")
    console.print(md)
    step_begin = datetime.now()
    res_atoms = create_hydrogen_reduced_structure(jobid=name, molecule=molecule)
    if res_atoms is not None:
        results.atoms = res_atoms
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _create_all_transition_states(console: Console, results: Results, name: str) -> None:
    #######################################################################
    # create all transition states
    #######################################################################
    md = Markdown(" Dock compound into transition-state templates")
    console.print(md)

    # define templates of the transition states
    # results.templates = [1, 4, 7, 19, 22, 25]
    step_begin = datetime.now()
    for i, atom in enumerate(results.atoms):
        create_transition_states(
            atoms=results.atoms, index=i, jobid=name, templates=results.templates
        )
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _create_csv_report(
    console: Console,
    data: Dict[str, List[float]],
    name: str,
    decimal_char: str = constant.CSV_DECIMAL_CHAR,
    index: bool = False,
    sep_char: str = constant.CSV_SEP_CHAR,
) -> Path:
    csv_file_path = Path.cwd() / f"{name}.csv"
    dataframe = pandas.DataFrame(data)
    with csv_file_path.open(mode='wt') as f:
        dataframe.to_csv(path_or_buf=f, decimal=decimal_char, index=index, sep=sep_char)

    console.print(f"Created a CSV file, {csv_file_path.name}, with relevant data.")

    return csv_file_path


def _create_molecular_image_report(
    console: Console,
    results: Results,
    name: str,
    molecule: T_RDKitMolecule,
) -> Path:
    #######################################################################
    # Create molecular image report, with calculated Boltzmann weights
    #######################################################################
    img_file_path = Path.cwd() / f"{name}.svg"
    create_boltzmann_weight_image(
        molecule=molecule,
        weights=results.weights,
        out_file_path=img_file_path,
    )

    console.print(
        f"Created molecular image, {img_file_path.name}, with Boltzmann weights."
    )

    return img_file_path


def _optimize_substrate_with_gfn2_xtb_alpb_thf(
    console: Console, results: Results, name: str, num_cpu_cores: int, xtb_opt_level: str
) -> None:
    #######################################################################
    # optimize substrate with GFN2-xtb/ALPB(THF)
    #######################################################################
    md = Markdown(" Optimize compound using GFN2-xtb/ALP(THF)")
    console.print(md)
    step_begin = datetime.now()
    path = Path.cwd() / f"{name}"
    file_path = path / "smiles.xyz"
    xtb_args = ["--opt", xtb_opt_level, "--alpb", "thf", "-P", f"{num_cpu_cores}"]
    res_substrate_energy = perform_xtb_calculation(
        input_file_path=file_path, args=xtb_args, dir_path=path
    )
    if res_substrate_energy is not None:
        results.substrate_energy = res_substrate_energy
    step_end = datetime.now()
    console.print("time: ", step_end - step_begin)


def _optimize_transition_states_with_gfn2_xtb_alpb_thf(
    console: Console, results: Results, name: str, num_cpu_cores: int, xtb_opt_level: str
) -> None:
    #######################################################################
    # GFN2-xtb/ALPB(THF) optimizations of transition states
    #######################################################################
    md = Markdown(" Optimize transition states with GFN2-xtb/ALPB(THF)")
    console.print(md)
    step_begin = datetime.now()
    xtb_args = [
        "--opt",
        xtb_opt_level,
        "--alpb",
        "thf",
        "--input",
        "constrain.inp",
        "-P",
        f"{num_cpu_cores}",
    ]
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        for template in results.templates:
            template_dir_path = base_dir_path / f"{template}"
            file_path = template_dir_path / "newstructure.xyz"
            xtb_begin = datetime.now()
            tmp = perform_xtb_calculation(
                input_file_path=Path(file_path), args=xtb_args, dir_path=template_dir_path
            )
            xtb_end = datetime.now()
            if tmp != float(14):
                print(f" - opt 1: {atom} - {template} {xtb_end - xtb_begin}")
            else:
                print(f" - opt 1: {atom} - {template} failed")

    # optimize everything and fix a dihedral
    xtb_args = [
        "--opt",
        xtb_opt_level,
        "--alpb",
        "thf",
        "--input",
        "fix.inp",
        "-P",
        f"{num_cpu_cores}",
    ]
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        for template in results.templates:
            template_dir_path = base_dir_path / f"{template}"
            file_path = template_dir_path / "xtbopt.xyz"
            xtb_begin = datetime.now()
            key = f"run-{atom}-{template}-2"
            results.xtb_energy[key] = perform_xtb_calculation(
                input_file_path=Path(file_path), args=xtb_args, dir_path=template_dir_path
            )
            xtb_end = datetime.now()
            if results.xtb_energy[key] != float(14):
                print(f" - opt 2: {atom} - {template} {xtb_end - xtb_begin}")
            else:
                print(f" - opt 2: {atom} - {template} failed")

    # HL gaps 1
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        for template in results.templates:
            template_dir_path = base_dir_path / f"{template}"
            file_path = template_dir_path / "xtb.out"
            key = f"run-{atom}-{template}-2"
            results.xtb_gaps[key] = xtb_helper.extract_homo_lumo_gap(file_path)

    # check if GFN2-xtb optimized structure is okay
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        results.xtb_ignore[atom] = []
        for template in results.templates:
            template_dir_path = base_dir_path / f"{template}"
            file_path = template_dir_path / "xtbopt.xyz"
            chem_helper.extract_substrate(file_path, 85, template_dir_path)
            ttpath = template_dir_path / "submolecule.xyz"
            try:
                kallisto_molecule = kallisto_helper.construct_molecule(geometry=ttpath)
                graph = kallisto_molecule.get_bonds()
                for j, node in enumerate(graph):
                    if sorted(node) == results.substrate_graph[atom][j]:
                        continue
                    else:
                        results.xtb_ignore[atom].append(template)
            except Exception:
                results.xtb_ignore[atom].append(template)

    # optimize substrate and fix catalyst (rotated)
    xtb_args = [
        "--opt",
        xtb_opt_level,
        "--alpb",
        "thf",
        "--input",
        "constrain.inp",
        "-P",
        f"{num_cpu_cores}",
    ]
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        for i_template in results.templates:
            s_template = f"{i_template}r"
            template_dir_path = base_dir_path / s_template
            file_path = template_dir_path / "newstructure.xyz"
            xtb_begin = datetime.now()
            tmp = perform_xtb_calculation(
                input_file_path=Path(file_path), args=xtb_args, dir_path=template_dir_path
            )
            xtb_end = datetime.now()
            if tmp != float(14):
                print(f" - opt 3: {atom} - {s_template} {xtb_end - xtb_begin}")
            else:
                print(f" - opt 3: {atom} - {s_template} failed")

    # optimize everything and fix a dihedral (rotated)
    xtb_args = [
        "--opt",
        xtb_opt_level,
        "--alpb",
        "thf",
        "--input",
        "fix.inp",
        "-P",
        f"{num_cpu_cores}",
    ]
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        for i_template in results.templates:
            s_template = f"{i_template}r"
            template_dir_path = base_dir_path / s_template
            file_path = template_dir_path / "xtbopt.xyz"
            key = f"run-{atom}-{s_template}-4"
            xtb_begin = datetime.now()
            results.xtb_energy[key] = perform_xtb_calculation(
                input_file_path=Path(file_path), args=xtb_args, dir_path=template_dir_path
            )
            xtb_end = datetime.now()
            if results.xtb_energy[key] != float(14):
                print(f" - opt 4: {atom} - {s_template} {xtb_end - xtb_begin}")
            else:
                print(f" - opt 4: {atom} - {s_template} failed")

    # HL gaps 2
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        for i_template in results.templates:
            s_template = f"{i_template}r"
            template_dir_path = base_dir_path / s_template
            key = f"run-{atom}-{s_template}-4"
            results.xtb_gaps[key] = xtb_helper.extract_homo_lumo_gap(
                template_dir_path / "xtb.out"
            )

    # check if GFN2-xtb optimized structure is okay
    kallisto_molecule = None
    for i, atom in enumerate(results.atoms):
        base_dir_path = Path.cwd() / f"{name}" / f"{atom}"
        results.xtb_ignore[atom] = []
        for i_template in results.templates:
            s_template = f"{i_template}r"
            template_dir_path = base_dir_path / s_template
            file_path = template_dir_path / "xtbopt.xyz"
            chem_helper.extract_substrate(file_path, 85, template_dir_path)
            ttpath = template_dir_path / "submolecule.xyz"
            if utility.exist_file(ttpath):
                kallisto_molecule = kallisto_helper.construct_molecule(geometry=ttpath)
            else:
                results.xtb_ignore[atom].append(i_template)
            graph = kallisto_molecule.get_bonds()
            for j, node in enumerate(graph):
                if sorted(node) == results.substrate_graph[atom][j]:
                    continue
                else:
                    results.xtb_ignore[atom].append(i_template)
    # remove duplicates from ignore list
    for i_key, i_val in results.xtb_ignore.items():
        results.xtb_ignore[i_key] = sorted(set(i_val))
    step_end = datetime.now()
    console.print("Ignore those structures: ", results.xtb_ignore)
    console.print("time: ", step_end - step_begin)


def _sort_ch_positions(console: Console, results: Results, name: str) -> None:
    #######################################################################
    # sort CH positions
    #######################################################################
    md = Markdown(" Prepare all structures for docking (sorting)")
    console.print(md)

    # save sorted substrate molecular graph
    # important to check if TS optimization retrains the structure
    step_begin = datetime.now()
    for i, atom in enumerate(results.atoms):
        results.substrate_graph[atom] = sort_ch_with_breadth_first_search(
            atoms=results.atoms, index=i, jobid=name
        )
    step_end = datetime.now()
    console.print("substrate graphs: ", results.substrate_graph)
    console.print("time: ", step_end - step_begin)


def _print_properties_to_console(console: Console, results: Results) -> rich.table.Table:
    #######################################################################
    # Print all properties
    #######################################################################

    # create a results table including all C-H positions and details
    table = rich.table.Table(show_header=True, header_style="bold magenta")
    table.add_column("[bold]C-H[/bold]", style="dim", width=12, justify="center")
    table.add_column("[bold]Similarity[/bold]", style="dim", width=16, justify="center")
    table.add_column(
        "[bold]Neighbor penalty[/bold]", style="dim", width=16, justify="center"
    )
    table.add_column(
        "[bold]PLS-PF penalty[/bold]", style="dim", width=16, justify="center"
    )
    table.add_column("[bold]Barrier[/bold]", style="dim", width=14, justify="center")
    table.add_column("[bold]Weight [%][/bold]", style="dim", width=14, justify="center")

    # fill table with C-H information (rows)
    for atom in results.atoms:
        table.add_row(
            f"{atom}",
            f"{results.similarity[atom]:.2f}",
            f"{results.graph_deltas[atom]:.2f}",
            f"{results.ml_deltas[atom]:.2f}",
            f"{results.barriers[atom]:.2f}",
            f"{results.weights[atom]:.2f}",
        )
    console.print(table)
    console.print(f"GFN2-xtb minimum energy: {results.xtb_minimum:.2f} kJ/mol")
    console.print("All energies in kJ/mol. Boltzmann weights in percent.")

    return table


def _rich_table_to_pandas_dataframe(table: rich.table.Table) -> Dict[str, List[float]]:
    """
    Convert a rich table into a Python dictionary with any rich formatting removed.

    :param table:
    """

    table_data = {
        rich.text.Text.from_markup(f'{x.header}').plain: [
            float(rich.text.Text.from_markup(f'{y}').plain) for y in x.cells
        ]
        for x in table.columns
    }
    return table_data


def _validate_args(
    parser: ArgumentParser, args: Namespace
) -> Tuple[T_RDKitMolecule, str]:
    # sanity checks for INPUT_FILE, NAME and SMILES values
    if args.name == "":
        parser.error("NAME may not be empty")

    if args.input_file is None:
        if args.smiles:
            return (
                rdkit_helper.load_smiles_as_rdkit_molecule(smiles=args.smiles),
                args.smiles,
            )
        else:
            parser.error("SMILES must be a valid SMILES string.")

    if args.input_file:
        input_file_path = Path(args.input_file)
        if not input_file_path.exists():
            parser.error("INPUT_FILE must exist.")

        valid_suffixes = [".mol", ".sdf", ".smi", ".smiles"]
        if input_file_path.suffix.lower() not in valid_suffixes:
            parser.error(f"INPUT_FILE supported formats {valid_suffixes}.")

        rdkit_molecule = rdkit_helper.load_mdl_file_as_rdkit_molecule(
            file_path=input_file_path
        )
        smiles = rdkit_helper.convert_rdkit_molecule_to_smiles(molecule=rdkit_molecule)
        return rdkit_molecule, smiles

    raise ValueError(f"CLI arguments are not valid, {args}")


if __name__ == "__main__":
    import sys

    os.environ["OPENBLAS_NUM_THREADS"] = "1"
    LOG = logging.getLogger(__name__)

    sys.exit(main())


import logging
from typing import Dict, List, Tuple, cast

import numpy as np

# noinspection PyPackageRequirements

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_weights(
    atoms: List[int],
    barriers: Dict[int, float],
    percent: bool = False,
    temperature_kelvin: float = 298.15,
    ideal_gas_constant: float = 0.0083145,
) -> Dict[int, float]:
    """Create Boltzmann weights from barriers given in kJ/mol.
    temperature_kelvin: temperature is given in Kelvin (default=298.15 K).
    ideal_gas_constant: R, the ideal gas constant given in kJ/mol (default=0.0083145 J/(K mol)."""

    # initialize weights
    weights: Dict[int, float] = {}

    norm = 0
    for i, atom in enumerate(atoms):
        exp_term = np.exp(-barriers[atom] / (ideal_gas_constant * temperature_kelvin))
        weights[atom] = exp_term
        norm += exp_term

    # divide by norm to get weights
    for atom in atoms:
        weights[atom] /= norm

    if percent:
        for key, value in weights.items():
            weights[key] = value * 100

    return weights


def similarity_switch(similarity: float, steepness: float) -> float:
    """Return float depending on the entered similarity."""
    return cast(float, 1 - np.tanh(steepness * (1 - similarity)))


def create_barriers(
    atoms: List[int],
    minimum: float,
    xtb_deltas: Dict[str, float],
    ml_deltas: Dict[int, float],
    graph_deltas: Dict[int, float],
    similarity: Dict[int, float],
    molecular_similarity: float,
) -> Tuple[Dict[int, float], Dict[int, float]]:
    """Create barriers for C-H activation for each position."""

    barriers = {}

    for atom in atoms:
        sim = similarity[atom]
        ml = ml_deltas[atom]
        graph = graph_deltas[atom]

        # for unhindered Hydrogen's use ML barrier
        if graph != 0.0:
            theta = similarity_switch(sim, 8.0)
            om_theta = 1 - theta

            ml *= theta
            graph *= om_theta

        mixed = ml + graph
        barriers[atom] = mixed + minimum

    #  T = 298.15 K and ideal_gas_constant = 0.0083145 J/(K mol)
    weights = create_weights(atoms=atoms, barriers=barriers, percent=True)

    return barriers, weights


import logging
import os
import shutil
from pathlib import Path

from icb.helper.utility import exist_directory

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def mkdir_p(path: Path) -> None:
    """Simulate bash command 'mkdir -p'."""
    try:
        path.mkdir(parents=True)
    except OSError as error:
        if path.is_dir():
            pass
        else:
            raise error


def create_directory(path: Path) -> Path:
    """Create directory with the given path."""

    if not path.exists():
        mkdir_p(path)

    return path


def copy_file(source: Path, destination: Path) -> None:
    """Copy file from source (path) to destination (path)."""
    shutil.copy2(source, destination)


def change_directory(path: Path) -> None:
    """Change directory to path."""
    if exist_directory(path):
        os.chdir(path)


import logging
from pathlib import Path
from typing import Dict

from icb.helper import rdkit_helper
from icb.types import T_RDKitMolecule

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_boltzmann_weight_image(
    molecule: T_RDKitMolecule,
    weights: Dict[int, float],
    out_file_path: Path,
    print_weights: bool = True,
    threshold: int = 5,
) -> None:
    """Create a molecular SVG image with RDKit where the calculated
    Boltzmann weights are visible at the C-H positions."""

    # copy RDKit molecule
    molecule_copy = rdkit_helper.copy_rdkit_molecule(molecule=molecule)

    # set weights
    for atom in molecule_copy.GetAtoms():
        idx = atom.GetIdx()
        try:
            if weights[idx] >= threshold:
                lbl = "%.0f" % (weights[idx])
                if print_weights:
                    atom.SetProp("atomNote", lbl)
                else:
                    atom.SetProp("atomNote", str(idx))
        except Exception:
            atom.SetProp("atomNote", "")

    # high quality SVG
    d2d = rdkit_helper.draw_rdkit_molecule_as_svg(
        molecule=molecule_copy, width=400, height=400
    )
    with out_file_path.open(mode="w") as f:
        f.write(d2d.GetDrawingText())


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['icb',
 'icb.helper',
 'icb.helper.db',
 'icb.helper.ml',
 'icb.helper.ts',
 'icb.main',
 'icb.task']

package_data = \
{'': ['*'], 'icb': ['data/*']}

install_requires = \
['click==8.1.3',
 'commonmark==0.9.1',
 'joblib==1.2.0',
 'kallisto==1.0.9',
 'pandas==1.5.1',
 'pillow==9.3.0',
 'pygments==2.13.0',
 'python-dateutil==2.8.2',
 'pytz==2022.6',
 'rdkit==2022.9.1',
 'rich==12.6.0',
 'scikit-learn==1.1.1',
 'scipy==1.9.2',
 'six==1.16.0',
 'threadpoolctl==3.1.0']

entry_points = \
{'console_scripts': ['sobo = icb.main.main:main']}

setup_kwargs = {
    'name': 'sobo',
    'version': '0.2.0',
    'description': 'Regioselectivity of the iridium-catalyzed borylation.',
    'long_description': '# Environment setup\nStart with setting up a new `conda` virtual environment (venv)\n\n```bash\n> conda create --name sobo python=3.9\n```\n\nand activate the created venv\n\n```bash\n> conda activate sobo\n```\n\n# Install dependencies\nGet `openbabel` and `xtb` dependencies\n\n```bash\n> conda install -c conda-forge openbabel\n...\n> conda install -c conda-forge xtb\n...\n```\n\nThen install the SoBo method\n\n```bash\n> pip install sobo\n```\n',
    'author': 'Caldeweyher, Eike',
    'author_email': 'e.caldeweyher@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
import statistics
from typing import Dict, List, Tuple, cast

import numpy as np

from icb.helper.xtb_helper import calculate_barrier

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def create_sterical_penalty(max_prox: float, prox: float) -> float:
    """Input a proximity shell as calculated from kallisto and return a sterical hindrance value in kJ/mol."""

    # logger = logging.getLogger("create_barriers")
    if max_prox < 12:
        max_prox = 12
    if LOG.isEnabledFor(logging.INFO):
        LOG.info(f"maxProx={max_prox}")
    return cast(float, 0.5 * (12 - 12 * np.tanh((max_prox - prox) / 2)))


def create_xtb_deltas(
    atoms: List[int],
    templates: List[int],
    substrate: float,
    xtb_energy: Dict[str, float],
    xtb_gaps: Dict[str, float],
    xtb_ignore: Dict[int, List[int]],
) -> Tuple[float, Dict[str, float]]:
    """Create deltas for all C-H positions using GFN2-xtb energies."""

    # catalyst = -120.568193063973 # opt-tight GFN2-xtb/ALPB(THF) energy in Hartree: catalyst
    catalyst = (
        -120.566060481458
    )  # opt-lax   GFN2-xtb/ALPB(THF) energy in Hartree: catalyst

    # errors for skipping certain templates
    errors = [
        "grep: opt2.xyz: No such file or directory",
    ]

    # dictionary for all barriers
    barriers: Dict[str, float] = {}
    barrier_list: Dict[str, List[float]] = {}
    for i, i_atom in enumerate(atoms):
        s_atom = f"{i_atom}"
        barrier_list[s_atom] = []

        tasks = []
        # start with normal substrate
        for i_template in templates:
            # if template in ignore cycle
            if not xtb_ignore[i_atom] and i_template in xtb_ignore[i_atom]:
                continue
            task = f"run-{i_atom}-{i_template}-2"
            tasks.append(task)
        for j, task in enumerate(tasks):
            # xtb optimization was not successful
            if xtb_energy[task] == 14:
                barrier = 14.0
            # HOMO-LUMO gap too small -> sort out
            if xtb_gaps[task] <= 0.1:
                continue
            barrier = calculate_barrier(
                ts=float(xtb_energy[task]),
                cat=float(catalyst),
                sub=float(substrate),
            )
            barrier_list[s_atom].append(barrier)

        tasks = []
        # continue with rotated substrate
        for i_template in templates:
            template_r = f"{i_template}r"
            # if template in ignore cycle
            if not xtb_ignore[i_atom] and template_r in xtb_ignore[i_atom]:
                continue
            task = f"run-{i_atom}-{template_r}-4"
            tasks.append(task)
        for j, task in enumerate(tasks):
            # xtb optimization was not successful
            if xtb_energy[task] == 14:
                barrier = 14.0
            # HOMO-LUMO gap too small -> sort out
            if xtb_gaps[task] <= 0.1:
                continue
            barrier = calculate_barrier(
                ts=float(xtb_energy[task]),
                cat=float(catalyst),
                sub=float(substrate),
            )
            barrier_list[s_atom].append(barrier)

    # create median barrier
    median: Dict[str, float] = {}
    for i, i_atom in enumerate(atoms):
        s_atom = f"{i_atom}"
        tmp = [b for b in barrier_list[s_atom] if b > 0]  # exclude negative barriers
        if len(tmp) == 0:
            median[s_atom] = 120
        else:
            median[s_atom] = statistics.median(tmp)

    # 14 kJ/mol represent a 300:1 Boltzmann weight
    sigma = 14

    # sort barriers out that are out of limit
    for i, i_atom in enumerate(atoms):
        s_atom = f"{i_atom}"
        # create limit from median barrier
        low_limit = median[s_atom] - sigma
        high_limit = median[s_atom] + 6 * sigma
        tmp = [
            b for b in barrier_list[s_atom] if low_limit <= b <= high_limit
        ]  # sort out barriers
        # in case there is nothing left :(
        if len(tmp) == 0:
            mean: float = 120
        else:
            mean = sum(tmp) / len(tmp)
        barriers[s_atom] = mean

    # get minimum barrier
    argmin = min(barriers, key=lambda x: barriers.get(x, float("nan")))
    minimum = barriers[argmin]

    # create GFN2-xtb deltas
    deltas = {k: v - minimum for k, v in barriers.items()}
    return minimum, deltas


import logging
import os
from pathlib import Path

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def exist_file(path: Path) -> bool:
    """Check if file with 'path' exists."""
    return path.exists() and path.is_file()


def exist_directory(path: Path) -> bool:
    """Check if directory with 'path' exists."""
    return path.exists() and path.is_dir()


def is_exe(path: Path) -> bool:
    """Check if program is available and executable."""
    return path.exists() and path.is_file() and os.access(path, os.X_OK)